package com.karneim.util.collection.regex;

import com.karneim.util.collection.set.ISet_char;
import com.karneim.util.collection.set.CharSet;

import java.io.Reader;
import java.io.IOException;

import java.util.*;

import com.karneim.util.collection.automaton.*;
import com.karneim.util.collection.set.AutomatonSet_String;

import java.text.ParsePosition;
import java.lang.ref.SoftReference;

/**
 * Regular expression based on a minimized deterministic automaton (FSA) and designed as a immutable set of strings.
 * <br>Use this class to create a regular expression and match strings against it.
 * <br>for example:
 * <br>to check whether a given string is a number try<br>
 * <br>new Pattern("[0-9]+").contains(s)
 *
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: Bro fr Softwarearchitektur   www.karneim.com</p>
 * @author Ralf Meyer
 * @version 1.0
 */
public class Pattern {

  private static final HashMap AUTOMATON_MAP = new HashMap();

  protected static final Automaton_Pattern get(String regEx) {
    synchronized(AUTOMATON_MAP) {
      final SoftReference reference = (SoftReference)AUTOMATON_MAP.get(regEx);

      if (reference!=null) {
        Automaton_Pattern automaton = (Automaton_Pattern)reference.get();
        if (automaton!=null) return automaton;
      }

      final Automaton_Pattern automaton = new Automaton_Pattern(regEx);

      final Automaton.LinkedSet_State states = automaton.getStates();
      for (Automaton.Wrapper_State w = states.elements; w!=null; w=w.next) {
        for (Automaton.State.Transition trans=w.state.eTransitions; trans!=null; trans=trans.next) {
          trans.properties = null;
        }
        for (Automaton.State.Transition trans=w.state.transitions; trans!=null; trans=trans.next) {
          trans.properties = null;
        }
      }

      automaton.minimize();

      AUTOMATON_MAP.put(regEx,new SoftReference(automaton));
      return automaton;
    }
  }


  protected Automaton_Pattern automaton;

  protected Pattern() {
    this.automaton = new Automaton_Pattern();
  }

  protected Pattern(Automaton_Pattern automaton) {
    this.automaton = automaton;
  }

  protected Pattern(ISet_char fullSet) {
    this.automaton = new Automaton_Pattern(fullSet);
  }

  public Pattern(String regEx) {
    this(Pattern.get(regEx));
  }

  public boolean contains(String s) {
    return this.contains(s,0,s.length());
  }

  public boolean contains(String s,int offset) {
    return this.contains(s,offset,s.length()-offset);
  }


  public boolean contains(String s,int offset,int length) {
    Automaton.State state = ((Automaton_Pattern)this.automaton).getStartState();
    if (state==null) return false;

    Automaton.LinkedSet_State states = this.automaton.newLinkedSet_State(state);
    Automaton.LinkedSet_State newStates = this.automaton.newLinkedSet_State();

    loop: for (;length>0; ++offset, --length) {
      for (Automaton.State.Transition trans=state.transitions; trans!=null; trans=trans.next) {
        if (trans.charSet.contains(s.charAt(offset))) {
          state = trans.toState;
          continue loop;
        }
      }
      return false;
    }

    return ((Automaton_Pattern.PState)state).isFinal();
  }

  public boolean contains(char[] chars) {
    return this.contains(chars,0,chars.length);
  }

  public boolean contains(char[] chars,int offset) {
    return this.contains(chars,offset,chars.length-offset);
  }


  public boolean contains(char[] chars,int offset,int length) {
    Automaton.State state = ((Automaton_Pattern)this.automaton).getStartState();
    if (state==null) return false;

    Automaton.LinkedSet_State states = this.automaton.newLinkedSet_State(state);
    Automaton.LinkedSet_State newStates = this.automaton.newLinkedSet_State();

    loop: for (;length>0; ++offset, --length) {
      for (Automaton.State.Transition trans=state.transitions; trans!=null; trans=trans.next) {
        if (trans.charSet.contains(chars[offset])) {
          state = trans.toState;
          continue loop;
        }
      }
      return false;
    }

    return ((Automaton_Pattern.PState)state).isFinal;
  }

  public boolean contains(java.io.Reader in) throws java.io.IOException {
    Automaton.State state = ((Automaton_Pattern)this.automaton).getStartState();
    if (state==null) return false;

    loop: for (int ch=in.read(); ch!=-1; ch=in.read()) {
      for (Automaton.State.Transition trans=state.transitions; trans!=null; trans=trans.next) {
        if (trans.charSet.contains((char)ch)) {
          state = trans.toState;
          continue loop;
        }
      }
      return false;
    }
    return ((Automaton_Pattern.PState)state).isFinal;
  }


  public String getRegEx() {
    return this.automaton.regEx;
  }

  public String toString() {
    return this.getRegEx();
  }

}


